﻿using Microsoft.Extensions.Logging;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using wclBluetooth;
using wclCommon;
using wclCommunication;

namespace BleService
{
    public class IotButton
    {
        #region Fields
        /// <summary>
        /// The UUID for the characteristic.
        /// </summary>
        private readonly Guid _uuidButtonPressed;
        /// <summary>
        /// The bluetooth radio.
        /// </summary>
        private readonly wclBluetoothRadio _radio;
        /// <summary>
        /// The UUID of the service.
        /// </summary>
        private readonly Guid _uuidButtonService;
        /// <summary>
        /// The ILogger of this class.
        /// </summary>
        private readonly ILogger _logger;
        /// <summary>
        /// The name of the device.
        /// </summary>
        private readonly string _name;
        /// <summary>
        /// The thread lock used to lock this thread until all events have been fired.
        /// </summary>
        private readonly SemaphoreSlim _signal;
        /// <summary>
        /// The gatt client used to connect to the device.
        /// </summary>
        private readonly wclGattClient _client;
        /// <summary>
        /// Used to hold all values.
        /// </summary>
        private readonly Dictionary<string, ushort> _values;		
		/// <summary>
		/// Field to hold the characteristic for button press.
		/// </summary>
        private wclGattCharacteristic characteristic;
        #endregion

        #region Properties
        /// <summary>
        /// Used to store the address of the device.
        /// </summary>
        public long Address { get; private set; }		
		/// <summary>
		/// Property to hold the characteristic for button press.
		/// </summary>
        public wclGattCharacteristic Characteristic
        {
            get
            {
                if (characteristic.IsNotifiable && characteristic.IsIndicatable)
                    characteristic.IsNotifiable = false;
                return characteristic;
            }
            private set
            {
                characteristic = value;
            }
        }
        #endregion

        #region Event Handlers
        /// <summary>
        /// This event handler is used to notify classes about a received value.
        /// </summary>
        public event EventHandler<ButtonPressEventArgs> OnValueReceived;
        #endregion

        #region CTOR
        /// <summary>
        /// The default constructor of this class.
        /// </summary>
        /// <param name="radio">The bluetooth radio.</param>
        /// <param name="address">The address of the device.</param>
        /// <param name="name">The name of the device.</param>
        /// <param name="uuidButtonPressed">The UUID of the characteristic.</param>
        /// <param name="uuidButtonService">THe UUD of the service.</param>
        /// <param name="logger">The ILogger of this class.</param>		
        public IotButton(wclBluetoothRadio radio, long address, string name, Guid uuidButtonPressed, Guid uuidButtonService, ILogger logger) : base()
        {
            _radio = radio;
            Address = address;
            _uuidButtonPressed = uuidButtonPressed;
            _uuidButtonService = uuidButtonService;
            _logger = logger;
            _name = name;
            _signal = new SemaphoreSlim(0, 1);
            _values = new Dictionary<string, ushort>();

            _client = new wclGattClient();
            _client.OnConnect += new wclClientConnectionConnectEvent(_client_OnConnect);
            _client.OnDisconnect += new wclClientConnectionDisconnectEvent(_client_OnDisconnect);
            _client.OnCharacteristicChanged += new wclGattCharacteristicChangedEvent(_client_OnCharacteristicChanged);
            _client.Address = Address;
        }
        #endregion

        #region Public Methods
        /// <summary>
        /// Initiates connecting, reading of values and disconnecting.
        /// </summary>
        public async Task Run()
        {
            var result = _client.Connect(_radio);
            if (result != wclErrors.WCL_E_SUCCESS)
                _logger.LogWarning($"Could not connect to bluetooth device with name {_name}. Error code {result.ToString("X8")}.");
            else
                await _signal.WaitAsync();
        }
        #endregion

        #region Private Methods
        /// <summary>
        /// Determines if the service is the right one, reads the characteristic and handles subscribe/unsubscribe/disconnect.
        /// </summary>
        private void ProcessService(wclGattService service)
        {
            if (service.Uuid.LongUuid == _uuidButtonService)
            {
                var characteristicsResult = _client.ReadCharacteristics(service, wclGattOperationFlag.goReadFromDevice, out var characteristics);
                if (characteristicsResult != wclErrors.WCL_E_SUCCESS)
                {
                    _logger.LogWarning($"Could not get characteristics from bluetooth device with name {_name}. Error code {characteristicsResult.ToString("X8")}.");
                    _client.Disconnect();
                }
                else
                {
                    _logger.LogDebug($"Got characteristics from bluetooth device with name {_name}");
                    foreach (var characteristic in characteristics)
                    {
                        if (characteristic.Uuid.LongUuid == _uuidButtonPressed)
                        {
                            Characteristic = characteristic;

                            if (Subscribe(Characteristic))
                                SubscribeWriteClientConfiguration(Characteristic);
							
							//TODO Wait for OnCharacteristicChanged to fire somehow...
							
                            UnsubscribeWriteClientConfiguration(Characteristic);
                            Unsubscribe(Characteristic);
                            SendShutdownSignal(Characteristic);
                        }
                    }
                }
            }
        }

		#region Private Methods
        /// <summary>
        /// Subscribes to a gatt characteristic.
        /// </summary>
        private bool Subscribe(wclGattCharacteristic characteristic)
        {
            var result = _client.Subscribe(characteristic);
            if (result != wclErrors.WCL_E_SUCCESS)
                _logger.LogWarning($"Could not subscribe to characteristic from device with name {_name}. Error code {result.ToString("X8")}.");
            else
                _logger.LogDebug($"Subscribed to characteristic from device with name {_name}");

            return result == wclErrors.WCL_E_SUCCESS;
        }

        /// <summary>
        /// Subscribes to a gatt write client configuration.
        /// </summary>
        private bool SubscribeWriteClientConfiguration(wclGattCharacteristic characteristic)
        {
            var result = _client.WriteClientConfiguration(characteristic, true, wclGattOperationFlag.goNone, wclGattProtectionLevel.plNone);

            if (result != wclErrors.WCL_E_SUCCESS)
                _logger.LogWarning($"Could not write client configuration for subscribe from device with name {_name}. Error code {result.ToString("X8")}.");
            else
                _logger.LogDebug($"Written client configuration for subscribe from device with name {_name}.");

            return result == wclErrors.WCL_E_SUCCESS;
        }

        /// <summary>
        /// Unsubscribes from a gatt characteristic.
        /// </summary>
        private bool Unsubscribe(wclGattCharacteristic characteristic)
        {
            var result = _client.Unsubscribe(characteristic);

            if (result != wclErrors.WCL_E_SUCCESS)
                _logger.LogWarning($"Could not unsubscribe from characteristic from device with name {_name}. Error code {result.ToString("X8")}.");
            else
                _logger.LogDebug($"Unsubscribed from characteristic from device with name {_name}");

            return result == wclErrors.WCL_E_SUCCESS;
        }

        /// <summary>
        /// Unsubscribes from a gatt write client configuration.
        /// </summary>
        private bool UnsubscribeWriteClientConfiguration(wclGattCharacteristic characteristic)
        {
            var result = _client.WriteClientConfiguration(characteristic, false, wclGattOperationFlag.goNone, wclGattProtectionLevel.plNone);

            if (result != wclErrors.WCL_E_SUCCESS)
                _logger.LogWarning($"Could not write client configuration for unsubscribe from device with name {_name}. Error code {result.ToString("X8")}.");
            else
                _logger.LogDebug($"Written client configuration for unsubscribe from device with name {_name}.");

            return result == wclErrors.WCL_E_SUCCESS;
        }

        /// <summary>
        /// Sends the shutdown signal to a characteristic so the client disconnects.
        /// </summary>
        private bool SendShutdownSignal(wclGattCharacteristic characteristic)
        {
            var disconnectCommand = Encoding.UTF8.GetBytes("1");
            var result = _client.WriteCharacteristicValue(characteristic, disconnectCommand);

            if (result == wclErrors.WCL_E_SUCCESS)
                _logger.LogDebug($"Successfully sent shutdown signal to device with name {_name}.");
            else
                _logger.LogWarning($"Could not send shutdown signal to device with name {_name}. Error code {result.ToString("X8")}.");

            return result == wclErrors.WCL_E_SUCCESS;
        }
        #endregion

        #region Events
        /// <summary>
        /// Fires when the client sends a new value.
        /// </summary>
        private void _client_OnCharacteristicChanged(object sender, ushort handle, byte[] value)
        {
            var strValue = Encoding.UTF8.GetString(value);

            var buttonId = strValue.Split(';')[0];
            var buttonPressCount = strValue.Split(';')[1];
            if (!_values.ContainsKey(buttonId))
            {
                _values.Add(buttonId, ushort.Parse(buttonPressCount));
                _logger.LogDebug($"Received new value from client: {strValue}");
            }
        }

        /// <summary>
        /// Fires when the gatt client has been disconnected.
        /// </summary>
        private void _client_OnDisconnect(object sender, int reason)
        {
            _logger.LogInformation($"Disconnected from bluetooth device with name {_name}.");
            OnValueReceived?.Invoke(this, new ButtonPressEventArgs(_values));
            _signal.Release();
        }

        /// <summary>
        /// Fires when the gatt client has been connected.
        /// </summary>
        private void _client_OnConnect(object sender, int error)
        {
            if (error != wclErrors.WCL_E_SUCCESS)
            {
                _logger.LogWarning($"Could not connect to bluetooth device with name {_name}. Error code {error.ToString("X8")}.");
                _signal.Release();
            }
            else
            {
                _logger.LogInformation($"Connected to bluetooth device with name {_name}.");
                var servicesResult = _client.ReadServices(wclGattOperationFlag.goNone, out var services);
                if (servicesResult != wclErrors.WCL_E_SUCCESS)
                {
                    _logger.LogWarning($"Could not get services from bluetooth device with name {_name}. Error code {servicesResult.ToString("X8")}.");
                    _client.Disconnect();
                }
                else
                {
                    _logger.LogDebug($"Got services from bluetooth device with name {_name}");
                    foreach (var service in services)
                    {
                        ProcessService(service);
                    }
                }
            }
        }
        #endregion
    }
}
